#include "HenonLike.h"

HenonLike::HenonLike(){
  loadCoef();
  latestArg.clear();
}

IVector HenonLike::latestArg;
interval HenonLike::thisY;
TCoefTwo HenonLike::sigma;
const interval HenonLike::B(1.9118611,1.9118725);
const interval HenonLike::C(0.82728796,0.82729138);
const interval HenonLike::E(0.2);
const interval HenonLike::unCertain(0.0000006*interval(-1,1));
const interval HenonLike::renormalisationScaling(interval(5.0)/5.13674200963507523);

void HenonLike::loadCoef() {
  int s,t;
  double in;
  std::ifstream coef("sigma.txt");
  
  for(int i=0; i<=230; i++) {
    coef>>s>>t>>in;
    sigma[s][t]=in;//*unCertain;
  }
  coef.close();
}

interval HenonLike::evalPhi(const TCoefOne &T, interval y) const{
  TCoefOne::const_iterator bg=T.begin(), en=T.end();
  interval result(0.0);
  --en;
  while(bg!=en) {
    result*=y;
    result+=(*en);
    --en;
  }
  result*=y;
  result+=(*en);
  return result;
}

interval HenonLike::evalDPhiDY(const TCoefOne &T, interval y) const{
  TCoefOne::const_iterator bg=T.begin(), en=T.end();
  interval result(0.0);
  --en;
  int exp(20);
  while(bg!=en) {
    result*=y;
    result+=exp*(*en);
    --en;
    --exp;
  }
 
  return result;
}

interval HenonLike::evalSigma(interval x, interval y) const{ 
  interval result=0;
  for(int i=20; i>=0; i--) {
    result*=x;
    TCoefOne tempVector=sigma.row(i);
    result+=evalPhi(tempVector,y); 
  }
  return result+unCertain;
}

interval HenonLike::evalS(interval x, interval y) const{
  return renormalisationScaling*
    ((C+E)*(x-1)+E*y+B*power(y,2)/2.0+evalSigma(x,y));
}

//Add Cauchy estimate
interval HenonLike::evalDSigmaDX(interval x, interval y) const {
  interval result=0;
  for(int i=20; i>=1; i--) {
    result*=x;
    TCoefOne tempVector=sigma.row(i);
    result+=i*evalPhi(tempVector,y); 
  }
  return result+unCertain/capd::abs(1.6-capd::abs(x-0.5));
}

//Add Cauchy estimate
interval HenonLike::evalDSigmaDY(interval x, interval y)  const{
  interval result=0;
  for(int i=20; i>=0; i--) {
    result*=x;
    TCoefOne tempVector=sigma.row(i);
    result+=evalDPhiDY(tempVector,y); 
  }
  return result+unCertain/capd::abs(1.6-capd::abs(y-0.5));
}


interval HenonLike::evalZeroFunc(interval x, interval y, interval u) const {
  return u+evalS(y,x);
}


interval HenonLike::evalY(interval x, interval u)  const{
  interval y0=mid((C+E*(1-x)-B*power(x,2)/2.0-u)/(C+E));
  y0+=mid(-evalZeroFunc(x, y0, u)/(C+E+evalDSigmaDX(y0,x)));
  //Now prove existence using interval newton !
 
  // interval inflatedY=interval(-1.1,2.1);
  interval inflatedY=interval(-1.0,2.0);
  if(!y0.subsetInterior(inflatedY))
    throw y0;
  interval Newton;
 
  int tooManyIt(0);
  do{
    Newton=y0-evalZeroFunc(x, y0, u)/(C+E+evalDSigmaDX(inflatedY,x));
    if (Newton.subsetInterior(inflatedY)) {
      intersection(Newton, inflatedY, inflatedY);
      break;
    }
    ++tooManyIt;
    if(tooManyIt>50)
      throw y0;
    y0=mid(Newton);
  }
  while( intersection(Newton, inflatedY, inflatedY) );

  int it(0);
 
  while (it<=5) {
  //std::cerr<<"it "<<std::endl;
    y0=mid(inflatedY);
    Newton=y0-evalZeroFunc(x, y0, u)/(C+E+evalDSigmaDX(inflatedY,x));
    intersection(Newton, inflatedY, inflatedY);
    ++it;
  }
  if(!intersection(interval(-1.1,2.1),inflatedY,inflatedY))
    throw y0;
  
  return inflatedY;
}

IVector HenonLike::rescale(IVector v, int k) const{
  //throw 1; //Use this when plotting unstable manifold
  std::cerr<<"RESCALING "<<std::endl;
  interval l(-0.24887681,-0.24887376);
  interval m(0.061107811,0.061112465);
  v[0]*=power(l,k);
  v[1]*=power(m,k);
  return v;
}

IVector HenonLike::evalF(const IVector &xu) const{
  interval y;
 
  if(!xu[0].subset(interval(-1.1,2.1))) {
    int k=1;
    while(!(power(-0.25,k)*xu[0]).subset(interval(-1.1,2.1)) ) {
      k++;

      //      if(std::abs(k)>1)
      //	throw k;
    }
    
    try{
      y=evalY(xu[0],xu[1]);
    }
    catch(...) {
      bool it;
      IVector temp=xu;
      do{
	it=false;
	temp=rescale(temp,1);
	try{
	  y=evalY(temp[0],temp[1]);
	}
	catch(...) {it=true; k++;}       
	//	if(std::abs(k)>1)
	// throw k;
      }
      while(it);
      
      return rescale(evalF(evalF(rescale(xu,k))),-k);
    }
  }
  else{
    try{
      y=evalY(xu[0],xu[1]);
    }
    catch(...) { // std::cout<<"error caught"<<std::endl; exit(0);
      int k=1;
      bool it;
      IVector temp=xu;
      do{
	it=false;
	temp=rescale(temp,1);
	try{
	  y=evalY(temp[0],temp[1]);
	}
	catch(...) {it=true;k++;}    
	//	if(std::abs(k)>1)
	// throw k;
      }
      while(it);

      return rescale(evalF(evalF(rescale(xu,k))),-k);
    }
  }
  IVector result;
  result[0]=y;
  result[1]=evalS(xu[0],y);
  latestArg=xu;
  thisY=y;
  return result;
}

IVector HenonLike::evalGradS(interval x, interval y)  const{
  IVector grad;
  grad[0]=(C+E)+evalDSigmaDX(x,y);
  grad[1]=E+B*y+evalDSigmaDY(x,y);
  return renormalisationScaling*grad;
}

IMatrix HenonLike::evalDF(interval x, interval y)  const{
  IMatrix DF;
  IVector gradS=evalGradS(y,x);
  DF[0][0]=-gradS[1]/gradS[0];
  DF[0][1]=-1.0/gradS[0];
  gradS=evalGradS(x,y);
  DF[1][0]=gradS[0]+gradS[1]*DF[0][0];
  DF[1][1]=gradS[1]*DF[0][1];
  return DF;
}

IVector HenonLike::operator()(const IVector &xu) const {
  return evalF(xu);
}

IMatrix HenonLike::operator[](const IVector &xu) const{
  if(xu==latestArg)
    return evalDF(xu[0],thisY);
 
  IVector temp=evalF(xu);
  return evalDF(xu[0],thisY);
}


IVector HenonLike::inv(IVector xu) const {
  xu[1]=-xu[1];
  xu=evalF(xu);
  xu[1]=-xu[1];
  return xu;
}
