#include "HenonLike.h"
#include <iomanip>
#include <fstream>
#include <stack>

/*
     Interval maximum finder

     Author:  Tomas Johnson <johnson@math.uu.se>
     Latest edit: Wed May 16 20:32:29 CEST 2007
     By: Tomas Johnson <johnson@math.uu.se>
*/

void findMaxMin(const interval &x, 
		const interval &y, 
		interval &uMin,
		interval &uMax,
		const HenonLike &H,
		bool first) {
  
  std::stack<interval> workList;
  workList.push(y);
  interval uMinLoc(1000);
  interval uMaxLoc(-1000);
  double tol(0.001);
  //std::cout<<x<<std::endl<<y<<std::endl;
  //findMin
  while(workList.size()!=0 ) {
    interval worky=workList.top();
    workList.pop();
    interval testS;
    bool fail(false);
    try {
      testS=-H.evalS(worky,x);
    }
    catch(...) {fail=true;}
    if(!fail && testS.leftBound()<=uMinLoc) {
      try {
	testS=-H.evalS(mid(worky),x);
      }
      catch(...) {fail=true;}
      if(!fail && testS.rightBound()<uMinLoc)
	uMinLoc=testS.rightBound();
    
      if(diam(worky)>tol) {
	workList.push(interval(worky.leftBound(),mid(worky).rightBound()));
	workList.push(interval(mid(worky).leftBound(), worky.rightBound()));
      }
    }
    else if(fail) {
      if(diam(worky)>tol) {
	workList.push(interval(worky.leftBound(),mid(worky).rightBound()));
	workList.push(interval(mid(worky).leftBound(), worky.rightBound()));
      }
    }
  }
  
  //findMax
  workList.push(y);
  while(workList.size()!=0 ) {
    interval worky=workList.top();
    workList.pop();
    interval testS;
    bool fail(false);
    try {
      testS=-H.evalS(worky,x);
    }
    catch(...) {fail=true;}
    if(!fail && testS.rightBound()>=uMaxLoc) {
      try {
	testS=-H.evalS(mid(worky),x);
      }
      catch(...) {fail=true;}
      if(!fail && testS.leftBound()>uMaxLoc)
	uMaxLoc=testS.leftBound();
      
      if(diam(worky)>tol) {
	workList.push(interval(worky.leftBound(),mid(worky).rightBound()));
	workList.push(interval(mid(worky).leftBound(), worky.rightBound()));
      }
    }
    else if(fail) {
      if(diam(worky)>tol) {
	workList.push(interval(worky.leftBound(),mid(worky).rightBound()));
	workList.push(interval(mid(worky).leftBound(), worky.rightBound()));
      }
    }
  }
  
  if(first) {
    uMin=uMinLoc;
    uMax=uMaxLoc;
  }
  else {
    if( uMinLoc>uMin)
      uMin=uMinLoc;
    if(uMaxLoc<uMax)
      uMax=uMaxLoc;
  }
  //std::cout<<uMinLoc<<std::endl
  //	   <<uMaxLoc<<std::endl<<std::endl;
}

bool isInside(double domain[][3], const IVector &im) {
  if(im[0].leftBound()<-1.1 || im[0].rightBound() >2.1)
    return false;
  int start, stop;
  interval uMin, uMax;
  bool started(false);
  for(int i=0; i<3200; i++) {
    if(!started && im[1].rightBound()>domain[i][0]) {
      started=true;
      start=i;
      uMin=domain[i][1];
      uMax=domain[i][2];
    }
    if(started) {
      if(im[1].leftBound()>domain[i][0]) {
	stop=i; break;
      }
      if(uMin < domain[i][1])
	uMin=domain[i][1];
      if(uMax > domain[i][2])
	uMax=domain[i][2];
    }
  }
  if(im[1].subsetInterior(interval(uMin.leftBound(),uMax.rightBound())))
    return true;
  else
    return false;
}

int main() {
  HenonLike H;
  int N=1000;
  /*
  interval uMin;
  interval uMax;
  findMaxMin(interval(-1.1,-1.1+1.0/(double)N),
	     interval(-1.1,2.1),uMin,uMax,H,true);
  
  std::ofstream Fout("FDomainFINEDONE.txt");
  Fout<<-1.1<<" "<<uMin.rightBound()<<" "<<uMax.leftBound()<<std::endl;
  for(int i=1; i<3.2*N; i++) {
    findMaxMin(interval(-1.1+i/(double)(N),-1.1+(i+1)/(double)(N)),
		interval(-1.1,2.1),uMin,uMax,H,true);
    Fout<<-1.1+i/(double)(N)<<" "<<uMin.rightBound()<<" "<<uMax.leftBound()<<std::endl;
  }
  Fout.close();
  */

  //REDO BELOW TOMORROW !!!
 
  std::ifstream Fin("FDomainD.txt");
  double domain[3201][3]; //3201
  //double domain[321][3];
  for(int i=0; i<3201; i++) {
    Fin>>domain[i][0]>>domain[i][1]>>domain[i][2];
  }
  Fin.close();
  std::ofstream Fout2("DomainTestInvStep2D.txt");
  /*  IVector tryBox;
  tryBox[0]=interval(-1.0,1.1);
  tryBox[1]=interval(-1.06,0.78);*/
  /* Domain[0]=interval(-1.1,2.1);
     Domain[1]=interval(-1.1,2.1);*/

  /*  for(int i=0; i<320; i++) {
    double h=leftBound((domain[i][2]-domain[i][1])/(double)(N+1.0));
    IVector tryB;
    tryB[0]=interval(domain[i][0],domain[i+1][0]);
    for(int j=0; j<N; j++) {
      tryB[1]=interval(domain[i][1]+j*h,domain[i][1]+(j+1.0)*h);
      IVector im=H(tryB);
      if(isInside(domain,im)) {
	im=H(im);
	if(isInside(domain,im)) {
	  Fout2<<tryB[0].leftBound()<<" "<<tryB[0].rightBound()<<" "
	       <<tryB[1].leftBound()<<" "<<tryB[1].rightBound()<<std::endl;
	}
      }
    }
  }
  Fout2.close();
  */
  /* for(int i=0; i<3200; i++) {
    double h=leftBound((domain[i][2]-domain[i][1])/(double)(N+1.0));
    IVector tryB;
    tryB[0]=interval(domain[i][0],domain[i+1][0]);
    for(int j=0; j<N; j++) {
      tryB[1]=interval(domain[i][1]+j*h,domain[i][1]+(j+1.0)*h);
      try {
	//	std::cout<<diam(tryB)<<std::endl;
	IVector im=H(midVector(tryB));
      	//std::cout<<diam(im)<<std::endl;
	if(isInside(domain,im)) {  //F2 well-def
	  im=H(midVector(im));
	  //std::cout<<diam(im)<<std::endl;
	  if(isInside(domain,im)) { //F3 well-def
	    im=H(midVector(im));
	    //std::cout<<diam(im)<<std::endl;
	    if(isInside(domain,im)) { //F4 well-def
	      im=H(midVector(im));
	      //std::cout<<diam(im)<<std::endl;
	      if(isInside(domain,im)) { //F5 well-def
		im=H(midVector(im));
		//std::cout<<diam(im)<<std::endl;
		if(isInside(domain,im)) { //F6 well-def 
		  //	std::cout<<tryB<<std::endl;
		  //tryB=H.rescale(tryB,-1);
		  //	std::cout<<tryB<<std::endl;
		  //  exit(0);
		  tryB[0]=tryB[0]/interval(-0.24887681,-0.24887376);
		  tryB[1]=tryB[1]/interval(0.061107811,0.061112465);
		  Fout2<<tryB[0].leftBound()<<" "<<tryB[0].rightBound()<<" "
		       <<tryB[1].leftBound()<<" "<<tryB[1].rightBound()<<std::endl;
		  // exit(0);
		}
	      }
	    }
	  }
	}
      } catch(...){ }
    }
  }
  Fout2.close();
  */
  /*
  for(int i=0; i<3200; i++) {
    double h=leftBound((domain[i][2]-domain[i][1])/(double)(N+1.0));
    IVector tryB;
    tryB[0]=interval(domain[i][0],domain[i+1][0]);
    for(int j=0; j<N; j++) {
      tryB[1]=interval(domain[i][1]+j*h,domain[i][1]+(j+1.0)*h);
      try {
	//	std::cout<<diam(tryB)<<std::endl;
	IVector im=H(tryB);
	//std::cout<<diam(im)<<std::endl;
	if(isInside(domain,im)) {  //F2 well-def
	  im=H(im);
	  //std::cout<<diam(im)<<std::endl;
	  if(isInside(domain,im)) { //F3 well-def
	    IVector tryInv;
	    tryInv=H.inv(tryB);
	    if(isInside(domain,tryInv)) { 
	      tryInv=H.inv(tryInv);
	      if(isInside(domain,tryInv)) { 
		tryInv=H.inv(tryInv);
		if(isInside(domain,tryInv)) { 
		  for(int k=0; k<4; k++) {
		    try {
		    IVector tryD;
		    if(k==0) {
		      tryD[0]=interval(tryInv[0].leftBound(),mid(tryInv[0]).rightBound());
		      tryD[1]=interval(tryInv[1].leftBound(),mid(tryInv[1]).rightBound());
		    }
		    else if(k==1) {
		      tryD[0]=interval(tryInv[0].leftBound(),mid(tryInv[0]).rightBound());
		      tryD[1]=interval(mid(tryInv[1]).leftBound(),tryInv[1].rightBound());
		    }
		    else if(k==2) {
		      tryD[0]=interval(mid(tryInv[0]).leftBound(),tryInv[0].rightBound());
		      tryD[1]=interval(tryInv[1].leftBound(),mid(tryInv[1]).rightBound());
		    }
		    else if(k==3) {
		      tryD[0]=interval(mid(tryInv[0]).leftBound(),tryInv[0].rightBound());
		      tryD[1]=interval(mid(tryInv[1]).leftBound(),tryInv[1].rightBound());
		    }
		    im=H(tryD);
		    if(isInside(domain,im)) {
		      im=H(im);
		      if(isInside(domain,im)) {
			im=H(im);
			if(isInside(domain,im)) {
			  tryD[0]=tryD[0]/interval(-0.24887681,-0.24887376);
			  tryD[1]=tryD[1]/interval(0.061107811,0.061112465);
			  Fout2<<tryD[0].leftBound()<<" "<<tryD[0].rightBound()<<" "
			       <<tryD[1].leftBound()<<" "<<tryD[1].rightBound()<<std::endl;		
			}
		      }
		    }
		    }
		    catch(...) { }
		  }
		}
	      }
	    }
	  }
	}
      } catch(...){ }
    }
  }
  Fout2.close();
  */

 
  //REDO THE ABOVE ONCE MORE
 for(int i=500; i<3200; i++) {
    double h=leftBound((domain[i][2]-domain[i][1])/(double)(N+1.0));
    IVector tryB;
    tryB[0]=interval(domain[i][0],domain[i+1][0]);
    for(int j=0; j<N; j++) {
      tryB[1]=interval(domain[i][1]+j*h,domain[i][1]+(j+1.0)*h);
      try {
	//	std::cout<<diam(tryB)<<std::endl;
	IVector im=H(tryB);
	//std::cout<<diam(im)<<std::endl;
	if(isInside(domain,im)) {  //F2 well-def
	  im=H(im);
	  //std::cout<<diam(im)<<std::endl;
	  if(isInside(domain,im)) { //F3 well-def
	    IVector tryInv;
	    tryInv=H.inv(tryB);
	    if(isInside(domain,tryInv)) { 
	      tryInv=H.inv(tryInv);
	      if(isInside(domain,tryInv)) { 
		tryInv=H.inv(tryInv);
		if(isInside(domain,tryInv)) { 
		  for(int k=0; k<4; k++) {
		    try {
		    IVector tryD;
		    if(k==0) {
		      tryD[0]=interval(tryInv[0].leftBound(),mid(tryInv[0]).rightBound());
		      tryD[1]=interval(tryInv[1].leftBound(),mid(tryInv[1]).rightBound());
		    }
		    else if(k==1) {
		      tryD[0]=interval(tryInv[0].leftBound(),mid(tryInv[0]).rightBound());
		      tryD[1]=interval(mid(tryInv[1]).leftBound(),tryInv[1].rightBound());
		    }
		    else if(k==2) {
		      tryD[0]=interval(mid(tryInv[0]).leftBound(),tryInv[0].rightBound());
		      tryD[1]=interval(tryInv[1].leftBound(),mid(tryInv[1]).rightBound());
		    }
		    else if(k==3) {
		      tryD[0]=interval(mid(tryInv[0]).leftBound(),tryInv[0].rightBound());
		      tryD[1]=interval(mid(tryInv[1]).leftBound(),tryInv[1].rightBound());
		    }
		    
		    tryD=H.inv(midVector(tryD));
		    if(isInside(domain,tryD)) { 
		      tryD=H.inv(tryD);
		      if(isInside(domain,tryD)) { 
			tryInv=H.inv(tryD);
			if(isInside(domain,tryD)) { 
			  for(int l=0; l<4; l++) {
			    IVector tryD2;
			    if(l==0) {
			      tryD2[0]=interval(tryD[0].leftBound(),mid(tryD[0]).rightBound());
			      tryD2[1]=interval(tryD[1].leftBound(),mid(tryD[1]).rightBound());
			    }
			    else if(l==1) {
			      tryD2[0]=interval(tryD[0].leftBound(),mid(tryD[0]).rightBound());
			      tryD2[1]=interval(mid(tryD[1]).leftBound(),tryD[1].rightBound());
			    }
			    else if(l==2) {
			      tryD2[0]=interval(mid(tryD[0]).leftBound(),tryD[0].rightBound());
			      tryD2[1]=interval(tryD[1].leftBound(),mid(tryD[1]).rightBound());
			    }
			    else if(l==3) {
			      tryD2[0]=interval(mid(tryD[0]).leftBound(),tryD[0].rightBound());
			      tryD2[1]=interval(mid(tryD[1]).leftBound(),tryD[1].rightBound());
			    }
			    tryD2=H.inv(midVector(tryD2));
			    if(isInside(domain,tryD2)) { 
			      tryD2=H.inv(tryD2);
			      if(isInside(domain,tryD2)) { 
				tryD2=H.inv(tryD2);
				if(isInside(domain,tryD2)) { 
				  for(int m=0; m<4; m++) {
				    IVector tryD3;
				    if(m==0) {
				      tryD3[0]=interval(tryD2[0].leftBound(),mid(tryD2[0]).rightBound());
				      tryD3[1]=interval(tryD2[1].leftBound(),mid(tryD2[1]).rightBound());
				    }
				    else if(m==1) {
				      tryD3[0]=interval(tryD2[0].leftBound(),mid(tryD2[0]).rightBound());
				      tryD3[1]=interval(mid(tryD2[1]).leftBound(),tryD2[1].rightBound());
				    }
				    else if(m==2) {
				      tryD3[0]=interval(mid(tryD2[0]).leftBound(),tryD2[0].rightBound());
				      tryD3[1]=interval(tryD2[1].leftBound(),mid(tryD2[1]).rightBound());
				    }
				    else if(m==3) {
				      tryD3[0]=interval(mid(tryD2[0]).leftBound(),tryD2[0].rightBound());
				      tryD3[1]=interval(mid(tryD2[1]).leftBound(),tryD2[1].rightBound());
				    }
				    
				    try {
				      for(int n=0; n<9; n++){
					tryD3=H(midVector(tryD3));
					if(!isInside(domain,tryD3)) throw im;
				      }
				      
				      tryD2[0]=tryD2[0]/power(interval(-0.24887681,-0.24887376),2);
				      tryD2[1]=tryD2[1]/power(interval(0.061107811,0.061112465),2);
				      Fout2<<tryD2[0].leftBound()<<" "<<tryD2[0].rightBound()<<" "
					   <<tryD2[1].leftBound()<<" "<<tryD2[1].rightBound()<<std::endl;
				      m=4; l=4; k=4;
				    }
				    catch(...) { }
				  }
				}
			      }
			    }
			  }
			}
		      }
		    }
		    
		    }
		    catch(...) { }
		  }
		}
	      }
	    }
	  }
	}
      } 
      catch(...){ }
    }
    std::cout<<i<<std::endl;
 }
 Fout2.close();
}
