#include "HenonLikealt.h"
#include <iomanip>
#include <fstream>


void initSets(HenonLike &H, HSet *hSet) {

  IVector centres[12][200];
  IVector wun;
  wun[0]= 0.788578889012330;
  wun[1]=-0.614933602760558;
 
  for(int i=0; i<200; i++) {
    centres[11][i][0]=0.577619;
    centres[11][i][1]=0.0;
    centres[10][i]=midVector(coord(centres[11][i]+0.00011*(i+1)/200.0*wun,false));
    centres[11][i]=coord(centres[11][i],false);
    for(int j=9; j>=0; j--) {
      centres[j][i]=midVector(H.inv(centres[j+1][i]));
      // std::cout<<centres[j][i]<<std::endl;
    }
    // std::cout<<centres[0][i]<<std::endl;
  }
  
  //  std::cout<<H[coord(centres[11][0],false)]<<std::endl;exit(0);
  //  exit(0);
  /* std::ofstream CENTRE("CENTRES_realFunction.txt", std::ios_base::out);
  CENTRE.precision(8);
  for(int i=0; i<200; i++) {
    for(int j=0; j<12; j++) {
      CENTRE<<centres[j][i]<<std::endl;
    }
  }
  exit(0);
  */
  
  
  std::cout<<"Generated covering sets"<<std::endl;
  IVector w;
  /*  w[0]=0.558987802533539;
      w[1]=-0.829175877976877;*/
  /* w[0]=0.546769451300831;
  w[1]=-0.837283206044519;
  */
  w[0]=0.570868623900820281;
  w[1]=-0.821041420541974730;

  double QF[]={1.0,-0.1}; //0.01 1 is OK  //unstable stable

  IVector wt;
  /*  wt[0]=0.997105803728644; 
      wt[1]=0.076026417583983;*/
  /*  wt[0]=-0.999597186245105;
      wt[1]=-0.028380719703166;*/
  wt[0]=0.992704972028756258;
  wt[1]=0.120568812341277774;

  //  double s=0.01;
  for(int i=0; i<12; i++) {
    for(int j=0; j<200; j++) {
      //  std::cout<<i<<" "<<j<<std::endl;
      hSet[i+j*12]=HSet(centres[i][j],0.001*w, 0.001*wt, QF); }// (centres+0.065*wt+0.015*w  //wt w
  }
  std::cout<<"init finished"<<std::endl;
}


void computeEdge(Frame &f1, Frame &f2,
                 FuncPointer f, const GridSet &G,
                 HSet &destination, int color, HenonLike &H, bool fixedPoint)
{
  unsigned i;
  for(i=0;i<G.center.size();i++)
    {
      IVector V = G.center[i] + G.C * G.r;
      
      f1.boxFill(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound(),color);
      
      f1.box(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound());
      if(fixedPoint) {
	V=H(V); //H(V)
      }
      else {
	V=H(H(V));
      }
      
      f2.boxFill(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound(),color);
      
      f2.box(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound());
      
      if(!f(destination,V)){
	rootFrame << At(4,0) << "Inclusion not satisfied";
	waitBt();
	exit(0);
      }
    }
} 

void setCoordinates(Frame &F, HSet &H) {
  double s=0.01;
  F.setWorldCoord(H.center[0]-s,H.center[1]-s,H.center[0]+s,H.center[1]+s);
}

void checkCoveringRelations(HSet *hSet, HenonLike &H) {
  GridSet G;
  int h=75;//  int h[] = {20,20,100}; //20, 20, 100
  int v=25;//int v[] = {20,20,100};
  bool flip[] = {true, true, true, true, true, true}; 
  Frame F1(50,50,450,450);
  Frame F2(550,50,950,450);
  for(int j=0; j<200; j++) {
    for(int i=0; i<12; i++) {
      rootFrame.Clear();
      F1.Bound(BROWN);
      F2.Bound(BROWN);
      int k=i+12*j;
      int l;
      i==11 ?  l=11+12*j : l=i+1+12*j;

      setCoordinates(F1, hSet[k]);
      setCoordinates(F2, hSet[l]);
      F1 << dynamic_cast<GtSet&> (hSet[k]);
      F2 << dynamic_cast<GtSet&> (hSet[l]);
      
      computeEdge(F1,F2,capd::covrel::Across,hSet[k].gridBottomEdge(G,h),hSet[l],GREEN,H, true);
      computeEdge(F1,F2,capd::covrel::Across,hSet[k].gridTopEdge(G,h),hSet[l],GREEN,H, true);
      computeEdge(F1,F2,capd::covrel::LeftSide,hSet[k].gridRightEdge(G,v),hSet[l],RED,H, true);
      computeEdge(F1,F2,capd::covrel::RightSide,hSet[k].gridLeftEdge(G,v),hSet[l],BLUE,H, true);
      
      std::cout<<k<<" OK"<<std::endl;
      // waitBt();
    }
  }
}

IMatrix gridDerivative(const HenonLike &H, GridSet &G, int it) {
  GridSet::iterator b=G.begin(), e=G.end();
  IMatrix DF;
  bool first(true);
  IVector box=G.C*G.r;
  while(b!=e) {
    IVector U=(*b)+box;
    IMatrix DFtemp=H[U];
    
    if(it==2) {
      U=H(U);
      DFtemp=H[U]*DFtemp;
    }
    /*if(it==3) {
      U=H(U);
      DFtemp=H[U]*DFtemp;
      U=H(U);
      DFtemp=H[U]*DFtemp;
    }
    */
    
    if(first) {
      first=false;
      DF=DFtemp;
    } 
    else {
      DF=intervalHull(DF,DFtemp);
    }
    ++b;
  }

  return DF;
}


void checkConeCondition(const HenonLike &H, HSet &source, HSet &destination, int unStableGrid, int stableGrid, int it) {
  GridSet G;
  source.gridSet(G,unStableGrid,stableGrid);

  IMatrix DF=gridDerivative(H,G,it);
  DF=capd::matrixAlgorithms::gaussInverseMatrix(destination.getCoordSystem())*DF*source.getCoordSystem();
 
  IMatrix V=Transpose(DF)*destination.get_QF()*DF-source.get_QF();
  if( !(V(1,1)>0 && (V(1,1)*V(2,2)-V(1,2)*V(2,1))>0 ) ) {
    std::cerr<<"Cone condition failed"<<std::endl
	     <<V(1,1)<<" "<<V(1,1)*V(2,2)-V(1,2)*V(2,1)<<std::endl;
    //exit(0);
  }
  //else {
    // std::cerr<<"Cone condition succeded"<<std::endl
    //     <<V(1,1)<<" "<<V(1,1)*V(2,2)-V(1,2)*V(2,1)<<std::endl;
  //}

}

void checkConeConditionsGlobally(HSet *hSet, const HenonLike &H) {
  for(int j=0; j<200; j++) {
    for(int i=0; i<11; i++) {
      checkConeCondition(H,hSet[i+12*j],hSet[i+1+12*j],50,50,1);
      std::cout<<"Cone Condition was checked for i = "<< i+12*j <<" -> i = "<< i+1+12*j <<std::endl;
    }
    checkConeCondition(H,hSet[11+12*j],hSet[11+12*j],50,50,1);
  }
 
}

int main(int argc, char* argv[])
{
  try{
    openGW(1000,1000, WHITE, BLACK);
    rootFrame.setWorldCoord(-2,-2,3,3);
    HSet hSet[2400];
    HenonLike H;
    initSets(H,hSet);
   
    checkCoveringRelations(hSet, H);
    checkConeConditionsGlobally(hSet, H);
  }
  catch(std::exception& e) {std::cout<<e.what()<<std::endl;}
  waitBt();
  closeGW();

}

