#include "HenonLike.h"
#include <iomanip>


void initSets(HSet *hSet) {
  IVector centres[4];
  IVector w[4];
  
  centres[0][0]=1.067744;
  centres[0][1]= 0.0;
  
  w[0][0]=0.271460881389316;
  w[0][1]=0.962449473933741;
  
  centres[1][0]= 0.90733000;
  centres[1][1]=-0.328353000000;
 
  w[1][0]= 0.653774812595729;
  w[1][1]=  0.691649449891044;
  
  centres[2][0]= 0.66401000;
  centres[2][1]=-0.0727302000000;
 
  w[2][0]=-0.736700425093906;
  w[2][1]= -0.676219257095255;

  
  centres[3][0]=interval(0.5776305);
  centres[3][1]=interval(0.0);
  w[3][0]=0.788578889012330;
  w[3][1]=0.614933602760558;


  IVector wt[4];
  for(int i=0; i<4; i++){
    wt[i][0]=w[i][0];
    wt[i][1]=-w[i][1];
  }
  double QF[]={1,-0.9};

  for(int i=0; i<3; i++) {
    hSet[i]=HSet(centres[i],0.01*wt[i],0.01*w[i], QF); //0.0034, 0.015 works...
  }

  hSet[3]=HSet(centres[3],0.035*wt[3],0.035*w[3], QF);
}


void computeEdge(Frame &f1, Frame &f2,
                 FuncPointer f, const GridSet &G,
                 HSet &destination, int color, HenonLike &H, bool fixedPoint)
{
  unsigned i;
  for(i=0;i<G.center.size();i++)
    {
      IVector V = G.center[i] + G.C * G.r;
      
      f1.boxFill(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound(),color);
      
      f1.box(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound());
      if(fixedPoint) {
	V=H(V);
      }
      else {
	V=H(H(V));
      }
      
      f2.boxFill(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound(),color);
      
      f2.box(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound());
      
      if(!f(destination,V)){
	rootFrame << At(4,0) << "Inclusion not satisfied";
	/*	waitBt();
	  exit(0);*/
      }
    }
} 

void setCoordinates(Frame &F, HSet &H) {
  double s=0.08; //0.08
  F.setWorldCoord(H.center[0]-s,H.center[1]-s,H.center[0]+s,H.center[1]+s);
}

void checkCoveringRelations(HSet *hSet, HenonLike &H) {
  GridSet G;
  int h[] = {20,20,100}; //20, 20, 100
  int v[] = {20,20,100};
  bool flip[] = {true, true, true, true, true, true}; 
  Frame F1(50,50,450,450);
  Frame F2(550,50,950,450);
  for(int i=0; i<3; i++) {
    rootFrame.Clear();
    F1.Bound(BROWN);
    F2.Bound(BROWN);
    
    setCoordinates(F1, hSet[i]);
    setCoordinates(F2, hSet[i+1]);
    F1 << dynamic_cast<GtSet&> (hSet[i]);
    F2 << dynamic_cast<GtSet&> (hSet[i+1]);
    if(flip[i]) {
      computeEdge(F1,F2,capd::covrel::Across,hSet[i].gridBottomEdge(G,h[i]),hSet[i+1],GREEN,H, false);
      computeEdge(F1,F2,capd::covrel::Across,hSet[i].gridTopEdge(G,h[i]),hSet[i+1],GREEN,H, false);
      computeEdge(F1,F2,capd::covrel::RightSide,hSet[i].gridLeftEdge(G,v[i]),hSet[i+1],RED,H, false);
      computeEdge(F1,F2,capd::covrel::LeftSide,hSet[i].gridRightEdge(G,v[i]),hSet[i+1],BLUE,H, false);
    }
    else {
      computeEdge(F1,F2,capd::covrel::Across,hSet[i].gridBottomEdge(G,h[i]),hSet[i+1],GREEN,H, false);
      computeEdge(F1,F2,capd::covrel::Across,hSet[i].gridTopEdge(G,h[i]),hSet[i+1],GREEN,H, false);
      computeEdge(F1,F2,capd::covrel::LeftSide,hSet[i].gridLeftEdge(G,v[i]),hSet[i+1],RED,H, false);
      computeEdge(F1,F2,capd::covrel::RightSide,hSet[i].gridRightEdge(G,v[i]),hSet[i+1],BLUE,H, false);
      
    }
    std::cerr<<i<<" OK"<<std::endl;
    waitBt();
  }
  rootFrame.Clear();
  setCoordinates(F1, hSet[3]);
  setCoordinates(F2, hSet[3]);
  F1<< dynamic_cast<GtSet&> (hSet[3]);
  F2<< dynamic_cast<GtSet&> (hSet[3]);
  computeEdge(F1,F2,capd::covrel::Across,hSet[3].gridBottomEdge(G,10),hSet[3],GREEN,H, true);
  computeEdge(F1,F2,capd::covrel::Across,hSet[3].gridTopEdge(G,10),hSet[3],GREEN,H, true);
  computeEdge(F1,F2,capd::covrel::LeftSide,hSet[3].gridRightEdge(G,10),hSet[3],RED,H, true);
  computeEdge(F1,F2,capd::covrel::RightSide,hSet[3].gridLeftEdge(G,10),hSet[3],BLUE,H, true);
  
}

IMatrix gridDerivative(const HenonLike &H, GridSet &G, int it) {
  GridSet::iterator b=G.begin(), e=G.end();
  IMatrix DF;
  bool first(true);
  IVector box=G.C*G.r;
  while(b!=e) {
    IVector U=(*b)+box;
    IMatrix DFtemp=H[U];
    if(it==2) {
      U=H(U);
      DFtemp=H[U]*DFtemp;
    }
    if(first) {
      first=false;
      DF=DFtemp;
    } 
    else {
      DF=intervalHull(DF,DFtemp);
    }
    ++b;
  }
  return DF;
}


void checkConeCondition(const HenonLike &H, HSet &source, HSet &destination, int unStableGrid, int stableGrid, int it) {
  GridSet G;
  source.gridSet(G,unStableGrid,stableGrid);

  IMatrix DF=gridDerivative(H,G,it);
  DF=capd::matrixAlgorithms::gaussInverseMatrix(destination.getCoordSystem())*DF*source.getCoordSystem();
 
  IMatrix V=Transpose(DF)*destination.get_QF()*DF-source.get_QF();
  if( !(V(1,1)>0 && (V(1,1)*V(2,2)-V(1,2)*V(2,1))>0 ) ) {
    std::cerr<<"Cone condition failed"<<std::endl
	     <<V(1,1)<<" "<<V(1,1)*V(2,2)-V(1,2)*V(2,1)<<std::endl;
    //exit(0);
  }

}

void checkConeConditionsGlobally(HSet *hSet, const HenonLike &H) {
  for(int i=0; i<3; i++) {
    checkConeCondition(H,hSet[i],hSet[i+1],6,6,2);
    std::cerr<<"Cone Condition is ok for i = " << i <<std::endl;
  }
  checkConeCondition(H,hSet[3],hSet[3],20,20,1);
  std::cerr<<"Cone Condition is ok for i = 3 -> 3 " <<std::endl;

}

void locateZero(const HenonLike &H) {
  IVector X;
  X[0]=0.577619;
  X[1]=0.0;
  std::cout<<H[X]<<std::endl; exit(0);
  /*X[0]=-2.32;
  X[1]=0.0;
  std::cout.precision(15);
  for(int i=0; i<20; i++) {
    X[0]+=mid((X[0]-H(X)[0])/(interval(-1.0)+H[X](1,1))); 
    std::cout<<X<<std::endl;
  }
  std::cout<<H[X]<<std::endl;*/
  /*  X[0]=1.067707;
  X[1]=0.0;
  X=H(H(X));
  std::cout<<X<<std::endl;
  X=H(H(X));
  std::cout<<X<<std::endl;
  X=H(H(X));
  std::cout<<X<<std::endl;
  X=H(H(X));
  std::cout<<X<<std::endl;
  */
  X[0]=-0.143;
  X[1]=0.0;
  std::cout.precision(15);
  for(int i=0; i<20; i++) {
    X[0]+=mid((X[0]-H(H(X))[0])/(interval(-1.0)+(H[H(X)]*H[X])(1,1))); 
    std::cout<<X<<std::endl;
  }
}

int main(int argc, char* argv[])
{

  try{
    openGW(1000,1000, WHITE, BLACK);
    rootFrame.setWorldCoord(-2,-2,3,3);
    HSet hSet[4];
    HenonLike H;
    initSets(hSet);
    /*    IVector temp;
    temp[0]=0.577619;
    temp[1]=0.0;
    for(int i=0; i<10; i++) {
      std::cout<<diam(temp)<<std::endl;
      temp=H(temp);
    }
    exit(0);*/
    // locateZero(H);
    //  exit(0);
    checkCoveringRelations(hSet, H);
    checkConeConditionsGlobally(hSet, H);
  }
  catch(std::exception& e) {std::cout<<e.what()<<std::endl;}
  waitBt();
  closeGW();

}

