#include "HenonLike.h"
#include <iomanip>
#include <fstream>


void initSets(HenonLike &H, HSet *hSet) {
  IVector centres[16];
 
  IVector w[8], wt[8];
  w[0][0]=0.788578889012330;
  w[0][1]= 0.614933602760558;
  w[1][0]= 0.982114476696540;
  w[1][1]= -0.188284770130463;
 
  interval l(-0.24887681,-0.24887376);
  interval m(0.061107811,0.061112465);

  IVector wr;
  wr[0]=w[1][0]/l;
  wr[1]=w[1][1]/m;
  interval nm=sqrt(power(wr[0],2)+power(wr[1],2));

  double QF[]={1.0,-0.8}; //0.01 1 is OK  //unstable stable

  wt[0][0]=w[0][0];
  wt[0][1]=-w[0][1];
  wt[1][0]=wr[0];
  wt[1][1]=-wr[1];

  centres[0][0]= 0.577619;
  centres[0][1]= 0.0;
 
  hSet[0]=HSet(centres[0],0.001*wt[0], 0.001*w[0], QF); 
  hSet[1]=HSet(centres[0],0.00072*wt[1], 0.00072*wr, QF); // stable unstable 

}


void computeEdge(Frame &f1, Frame &f2,
                 FuncPointer f, const GridSet &G,
                 HSet &destination, int color, HenonLike &H, bool notfixedPoint)
{
  unsigned i;
  for(i=0;i<G.center.size();i++)
    {
      IVector V = G.center[i] + G.C * G.r;
      
      f1.boxFill(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound(),color);
      
      f1.box(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound());
      if(notfixedPoint) {
	V=H(H(H(H(V)))); //H(V)
      }
      else {
	V=H(H(V));
      }
      
      f2.boxFill(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound(),color);
      
      f2.box(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound());
      
      if(!f(destination,V)){
	rootFrame << At(4,0) << "Inclusion not satisfied";
		std::cout<<"FAIL"<<std::endl;
		//	waitBt();
	//	exit(0);
      }
    }
} 

void setCoordinates(Frame &F, HSet &H) {
  double s=0.02;
  F.setWorldCoord(H.center[0]-s,H.center[1]-s,H.center[0]+s,H.center[1]+s);
}

void checkCoveringRelations(HSet *hSet, HenonLike &H) {
  GridSet G;
  int h=600;//600  int h[] = {20,20,100}; //20, 20, 100
  int v=500;//500 int v[] = {20,20,100};
  bool flip[] = {true, true, true, true, true, true}; 
  Frame F1(50,50,450,450);
  Frame F2(550,50,950,450);
  
  for(int i=2; i<3; i++) {
    rootFrame.Clear();
    F1.Bound(BROWN);
    F2.Bound(BROWN);
    int k,l;
    switch(i) {
    case 0:k=0; l=0; break;
    case 1:k=1; l=1; break;
    case 2:k=0; l=1; break;
    case 3:k=2; l=3; break;

    case 4:k=3; l=4; break;
     
    case 5:k=4; l=5; break; //

    case 6:k=5; l=6; break; //(

    case 7:k=6; l=7; break;
    case 8:k=7; l=8; break;
  
    case 9:k=8; l=9; break;
    case 10:k=9; l=10; break;
    case 11:k=10; l=0; break;
    }
 
    std::cout<<k<<" "<<l<<std::endl;
    setCoordinates(F1, hSet[k]);
    setCoordinates(F2, hSet[l]);
    F1 << dynamic_cast<GtSet&> (hSet[k]);
    F2 << dynamic_cast<GtSet&> (hSet[l]);
 
    computeEdge(F1,F2,capd::covrel::Across,hSet[k].gridBottomEdge(G,h),hSet[l],GREEN,H, false);
    computeEdge(F1,F2,capd::covrel::Across,hSet[k].gridTopEdge(G,h),hSet[l],GREEN,H, false);
    computeEdge(F1,F2,capd::covrel::RightSide,hSet[k].gridLeftEdge(G,v),hSet[l],RED,H, false);
    computeEdge(F1,F2,capd::covrel::LeftSide,hSet[k].gridRightEdge(G,v),hSet[l],BLUE,H, false);
    
    
    // waitBt();
  }
}

IMatrix gridDerivative(const HenonLike &H, GridSet &G, int it) {
  GridSet::iterator b=G.begin(), e=G.end();
  IMatrix DF;
  bool first(true);
  IVector box=G.C*G.r;
  int k=0;
  while(b!=e) {
    k++;
    if(k==500*(k/500))
      std::cerr<<k<<" ";
    IVector U=(*b)+box;
    IMatrix DFtemp=H[U];
    
    if(it==2) {
      U=H(U);
      DFtemp=H[U]*DFtemp;
    }
    if(it==3) {
      U=H(U);
      DFtemp=H[U]*DFtemp;
      U=H(U);
      DFtemp=H[U]*DFtemp;
    }
    if(it==4) {
      U=H(U);
      DFtemp=H[U]*DFtemp;
      U=H(U);
      DFtemp=H[U]*DFtemp;
      U=H(U);
      DFtemp=H[U]*DFtemp;
    }



    if(first) {
      first=false;
      DF=DFtemp;
    } 
    else {
      DF=intervalHull(DF,DFtemp);
    }
    ++b;
  }

  return DF;
}


void checkConeCondition(const HenonLike &H, HSet &source, HSet &destination, int unStableGrid, int stableGrid, int it) {
  GridSet G;
  source.gridSet(G,unStableGrid,stableGrid);

  IMatrix DF=gridDerivative(H,G,it);
  //  std::cout<<DF<<std::endl;
  DF=capd::matrixAlgorithms::gaussInverseMatrix(destination.getCoordSystem())*DF*source.getCoordSystem();
 
  IMatrix V=Transpose(DF)*destination.get_QF()*DF-source.get_QF();
  if( !(V(1,1)>0 && (V(1,1)*V(2,2)-V(1,2)*V(2,1))>0 ) ) {
    std::cerr<<"Cone condition failed"<<std::endl
	     <<V(1,1)<<" "<<V(1,1)*V(2,2)-V(1,2)*V(2,1)<<std::endl;
    //exit(0);
  }
  else  {
      std::cerr<<"Cone condition succeded"<<std::endl
	       <<V(1,1)<<" "<<V(1,1)*V(2,2)-V(1,2)*V(2,1)<<std::endl;
  }
  
}

void checkConeConditionsGlobally(HSet *hSet, const HenonLike &H) {
  checkConeCondition(H,hSet[0],hSet[1],50,50,2);
}


int main(int argc, char* argv[])
{
  //  int i=atoi(argv[1]);
  try{
    openGW(1000,1000, WHITE, BLACK);
    rootFrame.setWorldCoord(-2,-2,3,3);
    HSet hSet[16];
    HenonLike H;
    initSets(H,hSet);
    
    checkCoveringRelations(hSet, H);
    waitBt();
    closeGW();
    checkConeConditionsGlobally(hSet, H);
  }
  catch(std::exception& e) {std::cout<<e.what()<<std::endl;}

 

}

