#include "HenonLike.h"
#include <iomanip>
#include <fstream>


void initSets(HenonLike &H, HSet *hSet) {
  IVector centres[2];
 
  IVector w[2];
  w[0][0]=0.788578889012330;
  w[0][1]= 0.614933602760558;
  w[1][0]= 0.682259082166558;
  w[1][1]= -0.731107710380897;
 

  double QF[]={1.0,-0.8}; //0.01 1 is OK  //unstable stable

  IVector wt[2];
  wt[0][0]=w[0][0];
  wt[0][1]=-w[0][1];
  //wt[1][0]= w[1][0];
  //wt[1][1]=-w[1][1];
  wt[1][0]=0.750925931392967773;
  wt[1][1]=0.660386436536671957;

  centres[0][0]= 0.577619;
  centres[0][1]= 0.0;
  centres[1][0]= -0.52715552772576234;
  centres[1][1]= 0.0;
  centres[0]=centres[0]+0.0594*wt[0]+0.058*w[0];
  // centres[1]=centres[1]+0.065*wt[1]+0.0646*w[1]; 
  //  centres[1]=centres[1]+0.0645*wt[1]+0.0545*w[1]; 
  centres[1][0]=-0.441810513866242992;
  centres[1][1]=0.00304191772000443345;


  /*  std::cout<<centres[0]<<std::endl
	   <<centres[1]<<std::endl;
	   exit(0);*/
  /* hSet[0]=HSet(centres[0],0.0854*wt[0], 0.0854*w[0], QF); //  0.0854 0.0838
  //hSet[1]=HSet(centres[1],0.07*wt[1], 0.0648*w[1], QF); // stable unstable, 
  hSet[1]=HSet(centres[1],0.0655*wt[1], 0.0625*w[1], QF); // 0.0655 0.0625
  */
  /*WORKS WITH
0.0594 0.058
0.065 0.0646
0.0854 0.0838
0.07 0.0648
  */

  centres[0][1]=0.0;
  centres[1][1]=0.0;
  w[1][0]=wt[1][0];
  w[1][1]=-wt[1][1];

  hSet[0]=HSet(centres[0],0.083*wt[0], 0.083*w[0], QF); //  0.0854
  hSet[1]=HSet(centres[1],0.0655*wt[1], 0.0655*w[1], QF); //0.66
}


void computeEdge(Frame &f1, Frame &f2,
                 FuncPointer f, const GridSet &G,
                 HSet &destination, int color, HenonLike &H, bool fixedPoint)
{
  unsigned i;
  for(i=0;i<G.center.size();i++)
    {
      IVector V = G.center[i] + G.C * G.r;
      
      f1.boxFill(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound(),color);
      
      f1.box(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound());
      if(fixedPoint) {
	V=H(H(V)); //H(V)
      }
      else {
	V=H(H(H(V)));
      }
      
      f2.boxFill(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound(),color);
      
      f2.box(V[0].leftBound(),V[1].leftBound(),V[0].rightBound(),V[1].rightBound());
      
      if(!f(destination,V)){
	rootFrame << At(4,0) << "Inclusion not satisfied";
		std::cout<<"FAIL"<<std::endl;
	waitBt();
	//	exit(0);
      }
    }
} 

void setCoordinates(Frame &F, HSet &H) {
  double s=0.3;
  F.setWorldCoord(H.center[0]-s,H.center[1]-s,H.center[0]+s,H.center[1]+s);
}

void checkCoveringRelations(HSet *hSet, HenonLike &H) {
  GridSet G;
  int h=40000;//  int h[] = {20,20,100}; //20, 20, 100
  int v=1000;//int v[] = {20,20,100};
  bool flip[] = {true, true, true, true, true, true}; 
  Frame F1(50,50,450,450);
  Frame F2(550,50,950,450);
  
  for(int i=2; i<4; i++) {
    rootFrame.Clear();
    F1.Bound(BROWN);
    F2.Bound(BROWN);
    int k,l;
    switch(i) {
    case 0:k=0; l=0; break;
    case 1:k=0; l=1; break;
    case 2:k=1; l=1; break;
    case 3:k=1; l=0; break;
    }
 
    std::cout<<k<<" "<<l<<std::endl;
    setCoordinates(F1, hSet[k]);
    setCoordinates(F2, hSet[l]);
    F1 << dynamic_cast<GtSet&> (hSet[k]);
    F2 << dynamic_cast<GtSet&> (hSet[l]);
    
    computeEdge(F1,F2,capd::covrel::Across,hSet[k].gridBottomEdge(G,h),hSet[l],GREEN,H, false);
    computeEdge(F1,F2,capd::covrel::Across,hSet[k].gridTopEdge(G,h),hSet[l],GREEN,H, false);
    if(k==0 || k>=4) {
      computeEdge(F1,F2,capd::covrel::LeftSide,hSet[k].gridRightEdge(G,v),hSet[l],RED,H, false);
      computeEdge(F1,F2,capd::covrel::RightSide,hSet[k].gridLeftEdge(G,v),hSet[l],BLUE,H, false);
    }
    else {
      computeEdge(F1,F2,capd::covrel::RightSide,hSet[k].gridRightEdge(G,v),hSet[l],RED,H, false);
      computeEdge(F1,F2,capd::covrel::LeftSide,hSet[k].gridLeftEdge(G,v),hSet[l],BLUE,H, false);
    }
    
    waitBt();
  }
}

IMatrix gridDerivative(const HenonLike &H, GridSet &G, int it) {
  GridSet::iterator b=G.begin(), e=G.end();
  IMatrix DF;
  bool first(true);
  IVector box=G.C*G.r;
  int k=0;
  while(b!=e) {
    k++;
    if(k==500*(k/500))
      std::cerr<<k<<" ";
    IVector U=(*b)+box;
    IMatrix DFtemp=H[U];
    
    if(it==2) {
      U=H(U);
      DFtemp=H[U]*DFtemp;
    }
    if(it==3) {
      U=H(U);
      DFtemp=H[U]*DFtemp;
      U=H(U);
      DFtemp=H[U]*DFtemp;
    }
    
    if(first) {
      first=false;
      DF=DFtemp;
    } 
    else {
      DF=intervalHull(DF,DFtemp);
    }
    ++b;
  }

  return DF;
}


void checkConeCondition(const HenonLike &H, HSet &source, HSet &destination, int unStableGrid, int stableGrid, int it) {
  GridSet G;
  source.gridSet(G,unStableGrid,stableGrid);

  IMatrix DF=gridDerivative(H,G,it);
  //  std::cout<<DF<<std::endl;
  DF=capd::matrixAlgorithms::gaussInverseMatrix(destination.getCoordSystem())*DF*source.getCoordSystem();
 
  IMatrix V=Transpose(DF)*destination.get_QF()*DF-source.get_QF();
  if( !(V(1,1)>0 && (V(1,1)*V(2,2)-V(1,2)*V(2,1))>0 ) ) {
    std::cerr<<"Cone condition failed"<<std::endl
	     <<V(1,1)<<" "<<V(1,1)*V(2,2)-V(1,2)*V(2,1)<<std::endl;
    //exit(0);
  }
  else  {
      std::cerr<<"Cone condition succeded"<<std::endl
	       <<V(1,1)<<" "<<V(1,1)*V(2,2)-V(1,2)*V(2,1)<<std::endl;
  }
  
}

void checkConeConditionsGlobally(HSet *hSet, const HenonLike &H) {
  checkConeCondition(H,hSet[0],hSet[0],100,100,3);
  checkConeCondition(H,hSet[0],hSet[1],100,100,3);
  checkConeCondition(H,hSet[1],hSet[1],100,100,3);
  checkConeCondition(H,hSet[1],hSet[0],100,100,3);
}


int main(int argc, char* argv[])
{
  //  int i=atoi(argv[1]);
  try{
    openGW(1000,1000, WHITE, BLACK);
    rootFrame.setWorldCoord(-2,-2,3,3);
    HSet hSet[7];
    HenonLike H;
    initSets(H,hSet);
    
    checkCoveringRelations(hSet, H);
    waitBt();
    closeGW();
    //  checkConeConditionsGlobally(hSet, H);
  }
  catch(std::exception& e) {std::cout<<e.what()<<std::endl;}

 

}

